unit calendar_jp;

interface

uses NDate;

function HolidayName(const Date: TDate): string;
function Mansion(const Date: TDate): WideString;
function SexagenaryCycleDay(const Date: TDate): string;
function SolarTerm(const Date: TDate): string;

implementation

type
  THoliday = record
    From, To_, Month, Day, OrdWday, Wday: Integer;
    Func: function (const Date: TDate): Boolean;
    Name: string;
  end;

function Equinox(Year: Integer; RefDay: Double): Integer; 
const
  SolarYearDiff = 0.242194;
var
  dy: Integer;
begin
  dy := Year - 1980;
  if dy < 0 then
    Result := Trunc(RefDay + SolarYearDiff * dy - Trunc((dy-3) / 4))
  else
    Result := Trunc(RefDay + SolarYearDiff * dy - Trunc(dy / 4));
end;

function Syunbun(const Date: TDate): Boolean;
begin
  with Date do
    case Year of
      1900..1979: Result := Day=Equinox(Year, 20.8357);
      1980..2099: Result := Day=Equinox(Year, 20.8431);
    else
      Result := False;
    end;
end;

function Syuubun(const Date: TDate): Boolean;
begin
  with Date do
    case Year of
      1900..1979: Result := Day=Equinox(Year, 23.2588);
      1980..2099: Result := Day=Equinox(Year, 23.2488);
    else
      Result := False;
    end;
end;

const
  Holidays: array[0..32] of THoliday = (
    (From:1949;           Month: 1; Day: 1;            Name:''),
    (From:1949; To_:1999; Month: 1; Day:15;            Name:'l̓'),
    (From:2000;           Month: 1; OrdWday:2; Wday:0; Name:'l̓'),
    (From:1967;           Month: 2; Day:11;            Name:'LO̓'),
    (From:1949;           Month: 3; Func:Syunbun;      Name:'t̓'),
    (From:1949; To_:1988; Month: 4; Day:29;            Name:'Vca'),
    (From:1989; To_:2006; Month: 4; Day:29;            Name:'݂ǂ̓'),
    (From:2007;           Month: 4; Day:29;            Name:'a̓'),
    (From:1949;           Month: 5; Day: 3;            Name:'@LO'),
    (From:2007;           Month: 5; Day: 4;            Name:'݂ǂ̓'),
    (From:1949;           Month: 5; Day: 5;            Name:'ǂ̓'),
    (From:1996; To_:2002; Month: 7; Day:20;            Name:'C̓'),
    (From:2003; To_:2019; Month: 7; OrdWday:3; Wday:0; Name:'C̓'),
    (From:2020; To_:2020; Month: 7; Day:23;            Name:'C̓'),
    (From:2021;           Month: 7; OrdWday:3; Wday:0; Name:'C̓'),
    (From:2016; To_:2019; Month: 8; Day:11;            Name:'R̓'),
    (From:2020; To_:2020; Month: 8; Day:10;            Name:'R̓'),
    (From:2021;           Month: 8; Day:11;            Name:'R̓'),
    (From:1966; To_:2002; Month: 9; Day:15;            Name:'hV̓'),
    (From:2003;           Month: 9; OrdWday:3; Wday:0; Name:'hV̓'),
    (From:1949;           Month: 9; Func:Syuubun;      Name:'H̓'),
    (From:1966; To_:1999; Month:10; Day:10;            Name:'̈̓'),
    (From:2000; To_:2019; Month:10; OrdWday:2; Wday:0; Name:'̈̓'),
    (From:2020; To_:2020; Month: 7; Day:24;            Name:'X|[c̓'),
    (From:2021;           Month:10; OrdWday:2; Wday:0; Name:'X|[c̓'),
    (From:1948;           Month:11; Day: 3;            Name:'̓'),
    (From:1948;           Month:11; Day:23;            Name:'ΘJӂ̓'),
    (From:1989; To_:2018; Month:12; Day:23;            Name:'Vca'),
    (From:2020;           Month: 2; Day:23;            Name:'Vca'),

    (From:1959; To_:1959; Month: 4; Day:10; Name:'cqEmě̋V'),
    (From:1989; To_:1989; Month: 2; Day:24; Name:'aVc̑r̗'),
    (From:1990; To_:1990; Month:11; Day:12; Name:'ʂ̗琳a̋V'),
    (From:1993; To_:1993; Month: 6; Day: 9; Name:'cqEmě̋V')
    );

  SolarTermTab: array [1950..2050, 0..23] of Byte = (
    (6,21,4,19,6,21,5,21,6,22,6,22,8,23,8,24,8,23,9,24,8,23,8,22),
    (6,21,5,19,6,21,5,21,6,22,6,22,8,24,8,24,8,24,9,24,8,23,8,23),
    (6,21,5,20,6,21,5,20,5,21,6,21,7,23,7,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,8,24,8,22,7,22),
    (6,20,4,19,6,21,5,21,6,21,6,22,8,23,8,24,8,23,9,24,8,23,8,22),
    (6,21,4,19,6,21,5,21,6,22,6,22,8,24,8,24,8,24,9,24,8,23,8,23),
    (6,21,5,20,5,21,5,20,5,21,6,21,7,23,7,23,8,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,8,24,8,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,8,23,8,24,8,23,9,24,8,23,7,22),
    (6,21,4,19,6,21,5,21,6,22,6,22,8,23,8,24,8,24,9,24,8,23,8,22),
    (6,21,5,20,5,20,5,20,5,21,6,21,7,23,7,23,8,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,8,24,8,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,24,8,23,9,24,8,23,7,22),
    (6,21,4,19,6,21,5,21,6,22,6,22,8,23,8,24,8,24,9,24,8,23,8,22),
    (6,21,5,19,5,20,5,20,5,21,6,21,7,23,7,23,8,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,6,21,6,21,7,23,8,23,8,23,8,24,8,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,9,24,8,23,7,22),
    (6,21,4,19,6,21,5,21,6,22,6,22,8,23,8,24,8,24,9,24,8,23,8,22),
    (6,21,5,19,5,20,5,20,5,21,6,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,6,21,6,21,7,23,8,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,9,24,8,23,7,22),
    (6,21,4,19,6,21,5,21,6,22,6,22,8,23,8,24,8,24,9,24,8,23,8,22),
    (6,21,5,19,5,20,5,20,5,21,5,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,6,21,6,21,7,23,8,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,9,24,8,23,7,22),
    (6,21,4,19,6,21,5,21,6,22,6,22,8,23,8,24,8,24,9,24,8,23,8,22),
    (6,21,5,19,5,20,5,20,5,21,5,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,6,21,6,21,7,23,8,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,9,24,8,23,7,22),
    (6,21,4,19,6,21,5,21,6,22,6,22,8,23,8,24,8,24,9,24,8,23,8,22),
    (6,21,5,19,5,20,5,20,5,21,5,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,5,21,6,21,7,23,7,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,9,24,8,23,7,22),
    (6,21,4,19,6,21,5,21,6,22,6,22,8,23,8,24,8,23,9,24,8,23,8,22),
    (6,21,5,19,5,20,4,20,5,21,5,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,5,21,6,21,7,23,7,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,8,24,8,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,8,23,8,24,8,23,9,24,8,23,8,22),
    (6,21,4,19,5,20,4,20,5,21,5,21,7,22,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,5,21,5,20,5,21,6,21,7,23,7,23,8,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,8,24,8,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,24,8,23,9,24,8,23,7,22),
    (6,21,4,19,5,20,4,20,5,21,5,21,7,22,7,23,7,23,8,23,7,22,7,21),
    (5,20,4,19,5,20,5,20,5,21,6,21,7,23,7,23,8,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,6,21,6,21,7,23,8,23,8,23,8,24,8,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,9,24,8,23,7,22),
    (6,21,4,19,5,20,4,20,5,21,5,21,7,22,7,23,7,23,8,23,7,22,7,21),
    (5,20,4,18,5,20,5,20,5,21,6,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,6,21,6,21,7,23,8,23,8,23,8,23,8,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,9,24,8,23,7,22),
    (6,21,4,19,5,20,4,20,5,21,5,21,7,22,7,23,7,23,8,23,7,22,7,21),
    (5,20,4,18,5,20,5,20,5,21,5,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,6,21,6,21,7,23,8,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,9,24,8,23,7,22),
    (6,21,4,19,5,20,4,20,5,21,5,21,7,22,7,23,7,23,8,23,7,22,7,21),
    (5,20,4,18,5,20,5,20,5,21,5,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,6,21,6,21,7,23,8,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,9,24,8,23,7,22),
    (6,21,4,19,5,20,4,20,5,21,5,21,7,22,7,23,7,23,8,23,7,22,7,21),
    (5,20,4,18,5,20,5,20,5,21,5,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,5,21,6,21,7,23,7,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,9,24,8,23,7,22),
    (6,21,4,19,5,20,4,20,5,21,5,21,7,22,7,23,7,22,8,23,7,22,7,21),
    (5,20,4,18,5,20,5,20,5,21,5,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,5,21,6,21,7,23,7,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,8,24,8,23,7,22),
    (6,21,4,19,5,20,4,20,5,20,5,21,7,22,7,23,7,22,8,23,7,22,7,21),
    (5,20,4,18,5,20,4,20,5,21,5,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,5,21,6,21,7,23,7,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,22,7,23,8,23,8,23,8,24,8,22,7,22),
    (6,20,4,19,5,20,4,19,5,20,5,21,7,22,7,23,7,22,8,23,7,22,7,21),
    (5,20,3,18,5,20,4,20,5,21,5,21,7,22,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,5,21,5,20,5,21,6,21,7,23,7,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,6,21,5,20,6,21,6,21,7,23,8,23,8,23,8,24,8,22,7,22),
    (6,20,4,19,5,20,4,19,5,20,5,21,6,22,7,22,7,22,8,23,7,22,7,21),
    (5,20,3,18,5,20,4,20,5,21,5,21,7,22,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,5,20,5,20,5,21,6,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,6,21,6,21,7,23,8,23,8,23,8,24,8,22,7,22),
    (6,20,4,19,5,20,4,19,5,20,5,21,6,22,7,22,7,22,8,23,7,22,6,21),
    (5,20,3,18,5,20,4,20,5,21,5,21,7,22,7,23,7,23,8,23,7,22,7,21),
    (5,20,4,19,5,20,5,20,5,21,5,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,6,21,6,21,7,23,8,23,8,23,8,23,8,22,7,22),
    (6,20,4,19,5,20,4,19,5,20,5,21,6,22,7,22,7,22,8,23,7,22,6,21),
    (5,20,3,18,5,20,4,20,5,21,5,21,7,22,7,23,7,23,8,23,7,22,7,21),
    (5,20,4,18,5,20,5,20,5,21,5,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,6,21,6,21,7,23,8,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,5,20,4,19,5,20,5,21,6,22,7,22,7,22,8,23,7,22,6,21),
    (5,20,3,18,5,20,4,20,5,21,5,21,7,22,7,23,7,23,8,23,7,22,7,21),
    (5,20,4,18,5,20,5,20,5,21,5,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,6,21,6,21,7,23,8,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,5,20,4,19,5,20,5,21,6,22,7,22,7,22,8,23,7,22,6,21),
    (5,20,3,18,5,20,4,20,5,21,5,21,7,22,7,23,7,23,8,23,7,22,7,21),
    (5,20,4,18,5,20,5,20,5,21,5,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,5,21,6,21,7,23,7,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,5,20,4,19,5,20,5,21,6,22,7,22,7,22,8,23,7,22,6,21),
    (5,20,3,18,5,20,4,20,5,20,5,21,7,22,7,23,7,22,8,23,7,22,7,21),
    (5,20,4,18,5,20,4,20,5,21,5,21,7,23,7,23,7,23,8,23,7,22,7,22),
    (5,20,4,19,6,21,5,20,5,21,6,21,7,23,7,23,8,23,8,23,7,22,7,22),
    (6,20,4,19,5,20,4,19,5,20,5,21,6,22,7,22,7,22,7,23,7,22,6,21),
    (5,20,3,18,5,20,4,20,5,20,5,21,7,22,7,23,7,22,8,23,7,22,7,21),
    (5,20,4,18,5,20,4,20,5,21,5,21,7,22,7,23,7,23,8,23,7,22,7,22)
  );

  SolarTermNames: array[0..23] of string = (
    '', '劦', 't', 'J', '[', 't', '', 'J',
    '', '', '䊎', 'Ď', '', '友', 'H', '',
    'I', 'H', 'I', '~', '~', '', '', '~'
    );

  MansionNames: array[0..27] of WideString = (
    'p', '', WideChar($6c10), '[', 'S', '', '',
    'l', '', '', '', '', '', '',
    '', 'K', '', '', 'L', 'X', 'Q',
    '', 'S', '', '', '', '', 'f'
    );

  SexagenaryCycleNames: array[0..59] of string = (
    'bq','N','','K','C','Ȗ','M','h','p\','ᡓ',
    'b','','q','N','','ȉK','MC','h','p','ᡖ',
    'b\','','','','q','ȉN','M','hK','pC','ᡖ',
    'b','','\','','','Ȉ','Mq','hN','p','ᡉK',
    'bC','','','','\','ȓ','M','h','pq','ᡉN',
    'b','K','C','','','Ȗ','M\','h','p','ᡈ'
    );

function _HolidayName(const Date: TDate): string;
var
  hd: THoliday;
  i, wd: Integer;
  r: Boolean;
begin
  Result := '';
  r := False;
  wd := Date.WeekDay;
  for i := 0 to High(Holidays) do
  begin
    hd := Holidays[i];
    if Date.Year < hd.From then
      Continue;
    if (hd.To_ > 0) and (Date.Year > hd.To_) then
      Continue;
    if (hd.Month > 0) and (Date.Month <> hd.Month) then
      Continue;
    if (hd.Day > 0) and (Date.Day <> hd.Day) then
      Continue;

    if hd.OrdWday > 0 then
    begin
      if hd.Wday <> wd then
        Continue;
      case hd.OrdWDay of
        1: r := Date.Day <= 7;
        2: r := (Date.Day >= 8) and (Date.Day <= 14);
        3: r := (Date.Day >= 15) and (Date.Day <= 21);
        4: r := (Date.Day >= 22) and (Date.Day <= 28);
        5: r := Date.Day >= 29;
      end;
    end
    else if Assigned(hd.Func) then
      r := hd.Func(Date)
    else
      r := True;

    if r then
    begin
      Result := hd.Name;
      Break;
    end;
  end;
end;

function _IsHoliday(const Date: TDate): Boolean;
begin
  Result := _HolidayName(Date) <> '';
end;

function HolidayName(const Date: TDate): string;
var
  d, pd: TDate;
begin
  Result := _HolidayName(Date);

  if (Result = '') and (Date.Ord >= NDate.Date(1973, 4, 12).Ord) then
  begin
    d := Date;
    while True do
    begin
      pd := NDate.Date(d.Ord-1);
      if _IsHoliday(pd) then
      begin
        if pd.WeekDay = 6 then
        begin
          Result := 'U֋x';
          Break;
        end
      end else
        Break;
      d := pd;
    end;
  end;

  if (Result = '') and (Date.Ord >= NDate.Date(1985, 12, 17).Ord) then
    if (Date.WeekDay <> 6) and
        _IsHoliday(NDate.Date(Date.Ord-1)) and
        _IsHoliday(NDate.Date(Date.Ord+1)) then
      Result := '̋x';
end;

function Mansion(const Date: TDate): WideString;
begin
  Result := MansionNames[(Date.Ord + 24) mod 28];
end;

function SexagenaryCycleDay(const Date: TDate): string;
begin
  Result := SexagenaryCycleNames[(Date.Ord + 14) mod 60];
end;

function SolarTerm(const Date: TDate): string;
var
  d: Integer;
begin
  Result := '';
  if (Date.Year < Low(SolarTermTab)) or (Date.Year > High(SolarTermTab)) then
    Exit;
  d := (Date.Month-1) * 2;
  if Date.Day = SolarTermTab[Date.Year][d] then
    Result := SolarTermNames[d]
  else if Date.Day = SolarTermTab[Date.Year][d+1] then
    Result := SolarTermNames[d+1];
end;

end.
